#include "window.h"
#include "slideworld.h"

EXTERN int	TkConsoleInit (Tcl_Interp *interp);

//////////////////////////////////////////////////////////////////////
// Tcl/Tk Stuff
//

#ifdef WIN32
#define matherr _matherr
#else

// The following variable is a special hack that is needed in order for
// Sun shared libraries to be used for Tcl.
extern "C"
{
extern int matherr();
};

#endif // WIN32


int *tclDummyMathPtr = (int *) matherr;

// Called by Tk_Main() to let me init the modules (Togl) I will need.
int Tcl_AppInit(Tcl_Interp *pInterp);
VOID InitSLIDETCLGlobals(Tcl_Interp *pInterp);

//
// Tcl/Tk Stuff
//////////////////////////////////////////////////////////////////////

BOOL ParseCommandLine(UINT_32 &uiArgc, CHAR **ppcArgv);
VOID PrintUsage();
BOOL LoadWorld();

static CHAR gpcBuffer[200];
static CHAR *gpcProgramName = NULL;

int main(int argc, char **argv)
{
  UINT_32 uiArgc;
  CHAR **ppcArgv;

  // Make a copy of argc and argv so that we can take our 
  // arguments from it and modify it
  uiArgc = (UINT_32)argc;
  ppcArgv = new CHAR*[uiArgc];
  ASSERT( ppcArgv != NULL );
  memcpy(ppcArgv, argv, uiArgc*sizeof(CHAR*));

  // Parse the command line to find the input file name
  if ( !ParseCommandLine(uiArgc, ppcArgv) )
    {
      exit(1);
    }

  // This calls the event loop handler and only returns once the 
  // program is exited
  Tk_Main( (int)uiArgc, (char**)ppcArgv, Tcl_AppInit );

  return 0;
}

BOOL ParseCommandLine(UINT_32 &uiArgc, CHAR **ppcArgv)
{
  BOOL bCorrect;

  ASSERT( uiArgc > 0 );
  ASSERT( ppcArgv != NULL );

  bCorrect = TRUE;

  gpcProgramName = new CHAR[strlen(ppcArgv[0])+1];
  ASSERT( gpcProgramName != NULL );
  strcpy(gpcProgramName, ppcArgv[0]);

  if ( !(uiArgc == 1 || uiArgc == 3) )
    {
      PrintUsage();
      return FALSE;
    }

  //////////////////////////////////////////////////
  // Window UI Config File
  if ( uiArgc == 3 )
    {
      if ( strcmp(ppcArgv[1], "-f") == 0 )
	{
	  CWindow::SetUIConfigFile(ppcArgv[2]);
	}
      else
	{
	  PrintUsage();
	  return FALSE;
	}
    }

  return bCorrect;
}

VOID PrintUsage()
{
  fprintf(stderr, "Usage: %s [-f <window UI file>]\n", gpcProgramName);
}

BOOL LoadWorld()
{
  CSLIDEWorld *pCSLIDEWorld;
  CWindow *pCWindow;

  pCSLIDEWorld = new CSLIDEWorld;
  ASSERT( pCSLIDEWorld != NULL );
  pCSLIDEWorld->Init();

  pCWindow = new CWindow();
  ASSERT( pCWindow != NULL );
  pCWindow->Init("nofile.slf", 0, 0, "WINDOW");

  VERIFY(pCSLIDEWorld->InsertWindow(pCWindow));

  if ( !pCSLIDEWorld->CheckSemantics() )
    {
      fprintf(stderr, "Semantic Checking Done: %ld Errors and %ld Warnings.\n",
	      pCSLIDEWorld->GetErrorCount(), 
	      pCSLIDEWorld->GetWarningCount());
      fprintf(stderr, "Compilation halted because of semantic errors.\n");
      return FALSE;
    }

  CSLIDEWorld::SetWorld(pCSLIDEWorld);

  return TRUE;
}

int Tcl_AppInit(Tcl_Interp *pInterp)
{
  CSLIDEWorld::SetInterp(pInterp);

   // Initialize Tcl, Tk, and Oot.
   if (Tcl_Init(pInterp) == TCL_ERROR)
     {
       return TCL_ERROR;
     }
   if (Tk_Init(pInterp) == TCL_ERROR) 
     {
       return TCL_ERROR;
     }
  if (Togl_Init(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }

#ifdef WIN32
  if (TkConsoleInit(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }
#endif // WIN32

  if (Oot_Init(pInterp) == TCL_ERROR) 
    {
      return TCL_ERROR;
    }

  if ( CWindow::CreateOotClass(pInterp) == TCL_ERROR )
    {
       return TCL_ERROR;
    }
  Togl_CreateFunc( CWindow::CBHandleToglCreate );
  Togl_DisplayFunc( CWindow::CBHandleDisplay );
  Togl_ReshapeFunc( CWindow::CBHandleReshape );
  Togl_DestroyFunc( CWindow::CBHandleDestroy );

  InitSLIDETCLGlobals(pInterp);

  // Load the world from the file into memory
  if ( !LoadWorld() )
    {
      exit(2);
    }

  // This should Allocate all the windows for the application
  CSLIDEWorld::GetWorld()->Preprocess();

  return TCL_OK;
}

VOID InitSLIDETCLGlobals(Tcl_Interp *pInterp)
{
  UINT_32 i;

  ASSERT( pInterp != NULL );

  // Set up SLIDE Globals like SLF_TIME, SLF_MOUSE_X, SLF_MOUSE_Y
  Tcl_SetVar(pInterp, "SLF_PI", "3.1415927", TCL_GLOBAL_ONLY);
  Tcl_SetVar(pInterp, "SLF_TIME", "0", TCL_GLOBAL_ONLY);
  Tcl_SetVar(pInterp, "SLF_FRAME", "0", TCL_GLOBAL_ONLY);

  // On Off flags
  for ( i = 0; i < slideoftSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDEOnOffReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Solid flags
  for ( i = 0; i < slidestSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDESolidReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Shading flags
  gcppcSLIDEShadingReservedWords[slideshtSLF_INHERIT] = 
    gcppcSLIDEReservedWords[sliderwtSLF_INHERIT];
  for ( i = 0; i < slideshtSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDEShadingReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // LOD flags
  gcppcSLIDELODReservedWords[slidelodtSLF_OFF] = 
    gcppcSLIDEOnOffReservedWords[slideoftSLF_OFF];
  for ( i = 0; i < slidelodtSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDELODReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Projection flags
  for ( i = 0; i < slidepjtSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDEProjectionReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }

  // Lighting Type flags
  for ( i = 0; i < slideltSize; i++ )
    {
      sprintf(gpcBuffer, "%ld", i);
      Tcl_SetVar(pInterp, gcppcSLIDELightReservedWords[i], 
		 gpcBuffer, TCL_GLOBAL_ONLY);
    }
}

