#include "window.h"

#include "slideworld.h"

CWindow *gpCWindow = NULL;
CWindow *gpCWindowCurr = NULL;

static CHAR gpcBuffer[200];

static const INT_32 ciWindowSize = 500;

//////////////////////////////////////////////////////////////////////
// CWindow Class
//

Tcl_Interp *CWindow::m_pInterp = NULL;

CHashString *CWindow::m_pHashKeyboard = NULL;

const int CWindow::m_cuiOotMethods = cuiCWindowOotMethods;

// OOT Registration Variables
OotMethodEntry CWindow::m_pOotMethods[cuiCWindowOotMethods] =
{
  CWINDOW_METHODS
};
const CHAR *CWindow::m_cppcOotFields[CWindow::ftSize] =
{
  CWINDOW_FIELDS_STRINGS
};
const CHAR *CWindow::m_cpcOotClassName = "window";
OotClass *CWindow::m_pOotClass = NULL;

UINT_32 CWindow::m_uiOotCounter = 0;

CHAR *CWindow::m_pcUIConfigFile = NULL;

// Local Defaults
//static const BOOL cbDisplayEvents = TRUE;
static const BOOL cbDisplayEvents = FALSE;
//static const CWindow::EventLevelType ceEventLevel = CWindow::eltAction;
static const CWindow::EventLevelType ceEventLevel = CWindow::eltCallBack;

CWindow::CWindow()
{
  m_uiOotID = 0;
  m_pcOotName = NULL;
  m_pOotObject = NULL;

  m_pTogl = FALSE;
  m_bRedisplayPending = FALSE;

  m_pcFile = NULL;
  m_uiLine = 0;
  m_uiColumn = 0;
  m_pcSLIDEID = NULL;
}

CWindow::~CWindow()
{
  Uninit();
}

VOID CWindow::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		   CHAR *pcSLIDEID)
{
  UINT_32 i;

  ASSERT( pcFile != NULL );
  ASSERT( pcSLIDEID != NULL );

  // Set Up Keyboard Interpreter
  // This is only done once for the whole CWindow Class
  if ( m_pHashKeyboard == NULL )
    {
      m_pHashKeyboard = new CHashString;
      ASSERT( m_pHashKeyboard != NULL );
      m_pHashKeyboard->Init((3*gcuiglsiKeyBindings)/2);
      
      for ( i = 0; i < gcuiglsiKeyBindings; i++ )
	{
	  m_pHashKeyboard->Insert(&(gcpglsiKeyBindings[i]), gcpglsiKeyBindings[i].pcKey);
	}
    }

  m_uiOotID = 0;
  m_pcOotName = NULL;
  m_pOotObject = NULL;

  m_pTogl = FALSE;
  m_bRedisplayPending = FALSE;

  m_iWidth = ciWindowSize;
  m_iHeight = ciWindowSize;

  m_bDisplayEvents = cbDisplayEvents;
  m_eEventLevel = ceEventLevel;

  // SLIDE Stuff
  m_pcFile = pcFile;
  m_uiLine = uiLine;
  m_uiColumn = uiColumn;
  m_pcSLIDEID = pcSLIDEID;

  m_vOrigin.Null();
  m_vSize.Set(1.0, 1.0, 0.0);

  // Output Stuff
  m_pcOutputFile = NULL;
  m_fpOutputFile = NULL;

  // Set up the OOT communication link with Tcl
  InitTcl();

  // Create the actual Togl drawing area and initialize OpenGL
  //InitTogl();

  SetDisplayEvents(cbDisplayEvents);

  SetSecondsPerFrameRequested(CSLIDEWorld::GetSecondsPerFrameRequested());
  SetUpdateTime(0.0);
  SetRenderTime(0.0);  
}

VOID CWindow::InitTcl()
{
  //////////////////////////////////////////////////
  // Tcl Stuff
  m_uiOotID = (++m_uiOotCounter);
//  sprintf(gpcBuffer, "%s%ld", m_cpcOotClassName, m_uiOotID);
  sprintf(gpcBuffer, "slf%s", m_pcSLIDEID);
  m_pcOotName = new CHAR[strlen(gpcBuffer) + 1];
  ASSERT( m_pcOotName != NULL );
  strcpy(m_pcOotName, gpcBuffer);

  m_pOotObject = Oot_CreateObject(m_pInterp, m_pOotClass, m_pcOotName, this);
  ASSERT( m_pOotObject != NULL );
}

VOID CWindow::InitTogl()
{
  INT_32 iCode;

  ASSERT( gpCWindowCurr == NULL );
  gpCWindowCurr = this;

  // Create Togl Drawing Area
  sprintf(gpcBuffer, "frame .%s; togl .%s.toglarea -width 400 -height 400 -rgba true -double true -depth true -stencil true -ident %s", m_pcOotName, m_pcOotName, m_pcOotName);
  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );

  //printf("Created Togl Area\n");

  //////////////////////////////////////////////////
  // Window UI Config File
  if ( GetUIConfigFile() == NULL )
#ifdef WIN32
    {
      HKEY hkey, hkeyResult;
      LONG lResult;
      DWORD dwSize;
      ULONG ulType;
      
      lResult = RegOpenKeyEx(HKEY_CURRENT_USER, "Software", 0, KEY_READ, &hkeyResult);
      if ( lResult != ERROR_SUCCESS )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error reading registry string value: \\\"HKEY_CURRENT_USER\\\\Software\\\\UC Berkeley\\\\SLIDE\\\\1.0\\\\slidedraw\\\\UIConfigFile\\\"\\n Exiting program.\"");
	  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	  exit(1);
	}
      hkey = hkeyResult;
      
      lResult = RegOpenKeyEx(hkey, "UC Berkeley", 0, KEY_READ, &hkeyResult);
      if ( lResult != ERROR_SUCCESS )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error reading registry string value: \\\"HKEY_CURRENT_USER\\\\Software\\\\UC Berkeley\\\\SLIDE\\\\1.0\\\\slidedraw\\\\UIConfigFile\\\"\\n Exiting program.\"");
	  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	  exit(1);
	}
      ASSERT( lResult == ERROR_SUCCESS );
      hkey = hkeyResult;
      
      lResult = RegOpenKeyEx(hkey, "SLIDE", 0, KEY_READ, &hkeyResult);
      if ( lResult != ERROR_SUCCESS )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error reading registry string value: \\\"HKEY_CURRENT_USER\\\\Software\\\\UC Berkeley\\\\SLIDE\\\\1.0\\\\slidedraw\\\\UIConfigFile\\\"\\n Exiting program.\"");
	  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	  exit(1);
	}
      ASSERT( lResult == ERROR_SUCCESS );
      hkey = hkeyResult;

      lResult = RegOpenKeyEx(hkey, "1.0", 0, KEY_READ, &hkeyResult);
      if ( lResult != ERROR_SUCCESS )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error reading registry string value: \\\"HKEY_CURRENT_USER\\\\Software\\\\UC Berkeley\\\\SLIDE\\\\1.0\\\\slidedraw\\\\UIConfigFile\\\"\\n Exiting program.\"");
	  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	  exit(1);
	}
      ASSERT( lResult == ERROR_SUCCESS );
      hkey = hkeyResult;

      lResult = RegOpenKeyEx(hkey, "slidedraw", 0, KEY_READ, &hkeyResult);
      if ( lResult != ERROR_SUCCESS )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error reading registry string value: \\\"HKEY_CURRENT_USER\\\\Software\\\\UC Berkeley\\\\SLIDE\\\\1.0\\\\slidedraw\\\\UIConfigFile\\\"\\n Exiting program.\"");
	  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	  exit(1);
	}
      ASSERT( lResult == ERROR_SUCCESS );
      hkey = hkeyResult;

      dwSize = 200;
      lResult = RegQueryValueEx(hkeyResult, "UIConfigFile", 0, &ulType, (LPBYTE)gpcBuffer, &dwSize);
      if ( lResult != ERROR_SUCCESS || ulType != REG_SZ )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error reading registry string value: \\\"HKEY_CURRENT_USER\\\\Software\\\\UC Berkeley\\\\SLIDE\\\\1.0\\\\slidedraw\\\\UIConfigFile\\\"\\n Exiting program.\"");
	  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	  exit(1);
	}

      CWindow::SetUIConfigFile(gpcBuffer);
    }
#else // WIN32
    {
      SetUIConfigFile("/project/cs/sequin/caffe/slide/lib/slidedraw/window.tcl");
    }
#endif // WIN32

  ASSERT( m_pcUIConfigFile != NULL );
  iCode = Tcl_EvalFile(m_pInterp, m_pcUIConfigFile); 
  if ( iCode != TCL_OK )
    {
      sprintf(gpcBuffer, "tk_messageBox -icon error -type ok -default ok -message \"Error opening UI configure file: \\\"%s\\\"\\n Exiting program.\"", m_pcUIConfigFile);
      iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
      if ( iCode != TCL_OK )
	{
	  printf("MessageBox returned: \"%s\"\n", m_pInterp->result);
	}
      ASSERT( iCode == TCL_OK );
      exit(1);
    }

  //printf("Sourced CreateWindow Function\n");

  sprintf(gpcBuffer, "CreateWindow . %s", m_pcOotName);
  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );

  //printf("Created Window\n");

  sprintf(gpcBuffer, "pack .%s", m_pcOotName);
  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );

  //printf("Packed Window\n");
}

VOID CWindow::Uninit()
{
  if ( m_pcOotName != NULL )
    {
      delete []m_pcOotName;
      m_pcOotName = NULL;
    }

  // We do not own this memory
  m_pOotObject = NULL;
  m_pTogl = NULL;

  // Output File
  if ( m_pcOutputFile != NULL )
    {
      delete []m_pcOutputFile;
      m_pcOutputFile = NULL;
    }
  ASSERT( m_fpOutputFile == NULL );
}

BOOL CWindow::CheckSemantics(CSLIDEWorld * /*pCSLIDEWorld*/)
{
  return TRUE;
}

VOID CWindow::Preprocess()
{
  // Open the Window
  // Create the actual Togl drawing area and initialize OpenGL
  InitTogl();
}

VOID CWindow::Update()
{
  PostRedisplay();
}

VOID CWindow::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

///////////////////////////////////////////////////
// START: ASSIGNMENT 1: Write SLIDE Description
// 

  /////////////////////////////////////////////////
  // START: Replace this square with your edited object
  //

  fprintf(m_fpOutputFile, "point pSquareXY ( 0.75 0.75 0 )\n");
  fprintf(m_fpOutputFile, "endpoint\n");
  fprintf(m_fpOutputFile, "point pSquarexY ( 0.25 0.75 0 )\n");
  fprintf(m_fpOutputFile, "endpoint\n");
  fprintf(m_fpOutputFile, "point pSquarexy ( 0.25 0.25 0 )\n");
  fprintf(m_fpOutputFile, "endpoint\n");
  fprintf(m_fpOutputFile, "point pSquareXy ( 0.75 0.25 0 )\n");
  fprintf(m_fpOutputFile, "endpoint\n");
  fprintf(m_fpOutputFile, "\n");
  fprintf(m_fpOutputFile, "face fSquare ( pSquareXY pSquarexY pSquarexy pSquareXy )\n");
  fprintf(m_fpOutputFile, "endface\n");
  fprintf(m_fpOutputFile, "\n");
  fprintf(m_fpOutputFile, "object oSquare ( fSquare )\n");
  fprintf(m_fpOutputFile, "  solid SLF_HOLLOW\n");
  fprintf(m_fpOutputFile, "  shading SLF_WIRE\n");
  fprintf(m_fpOutputFile, "endobject\n");
  fprintf(m_fpOutputFile, "\n");

  //
  // END  : Replace this square with your edited object
  /////////////////////////////////////////////////

  // The following is the minimal viewing parameter setup which the renderer
  // needs to display your object

  fprintf(m_fpOutputFile, "camera cam\n");
  fprintf(m_fpOutputFile, "  projection SLF_PARALLEL\n");
  fprintf(m_fpOutputFile, "  frustum ( -0.5 -0.5 -2 )\n");
  fprintf(m_fpOutputFile, "          ( 0.5 0.5 -0.01 )\n");
  fprintf(m_fpOutputFile, "endcamera\n");
  fprintf(m_fpOutputFile, "group gCam\n");
  fprintf(m_fpOutputFile, "  instance cam\n");
  fprintf(m_fpOutputFile, "    id iCam\n");
  fprintf(m_fpOutputFile, "    translate ( 0.5 0.5 1 )\n");
  fprintf(m_fpOutputFile, "  endinstance\n");
  fprintf(m_fpOutputFile, "endgroup\n");
  fprintf(m_fpOutputFile, "\n");
  fprintf(m_fpOutputFile, "window win\n");
  fprintf(m_fpOutputFile, "endwindow\n");
  fprintf(m_fpOutputFile, "\n");
  fprintf(m_fpOutputFile, "viewport vp win\n");
  fprintf(m_fpOutputFile, "endviewport\n");
  fprintf(m_fpOutputFile, "\n");
  fprintf(m_fpOutputFile, "render vp gCam.iCam.cam oSquare\n");
  fprintf(m_fpOutputFile, "endrender\n");

// 
// END  : ASSIGNMENT 1: Write SLIDE Description
///////////////////////////////////////////////////
}

VOID CWindow::HandleReshape(INT_32 iWidth, INT_32 iHeight)
{
  DisplayEvent(eltCallBack, "Entering HandleReshape(%3ld, %3ld)\n", 
	       iWidth, iHeight);

  CListIterInt iter;

  m_iWidth = iWidth;
  m_iHeight = iHeight;

  gilViewport(0, 0, m_iWidth, m_iHeight);

  gilMatrixMode(GL_PROJECTION);
  gilLoadIdentity();
  gilOrtho(0, 1, 0, 1, -1, 1);

  gilMatrixMode(GL_MODELVIEW);
  gilLoadIdentity();
}

VOID CWindow::HandleDisplay()
{
  DisplayEvent(eltCallBack, "Entering HandleDisplay()\n");

  CListIterInt iter;
  Timer t;

  t.Reset();
  t.Start();

  m_bRedisplayPending = FALSE;

  // Make our Togl Window the current OpenGL Context
  ASSERT( m_pTogl != NULL );
  Togl_MakeCurrent(m_pTogl);

  gilClearColor(0.0, 0.0, 0.0, 0.0);
  gilClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  gilDisable(GL_DEPTH_TEST);
  gilShadeModel(GL_FLAT);
  gilDisable(GL_CULL_FACE);
  gilDisable(GL_TEXTURE_2D);
  gilDepthFunc(GL_LEQUAL);
  gilDisable(GL_LIGHTING);
  gilDisable(GL_NORMALIZE);

///////////////////////////////////////////////////
// START: ASSIGNMENT 1: Rendering Code
// 
  // Here is some sample polygon drawing code which 
  // you should replace with your drawing editor's 
  // rendering code.

  // Cross
  gilColor3f(1.0, 0.0, 0.0);

  gilBegin(GL_LINES);
  {
    gilVertex2f(0.5, 0.0);
    gilVertex2f(0.5, 1.0);

    gilVertex2f(0.0, 0.5);
    gilVertex2f(1.0, 0.5);
  }
  gilEnd();


  gilColor3f(0.0, 1.0, 1.0);

  // First Left Clicked Point
  gilPointSize(3.0);
  gilBegin(GL_POINTS);
  {
    gilVertex2f((FLOAT)0.05, (FLOAT)0.95);
  }
  gilEnd();

  // Many Left Clicked Points
  gilBegin(GL_LINE_STRIP);
  {
    gilVertex2f((FLOAT)0.55, (FLOAT)0.95);
    gilVertex2f((FLOAT)0.55, (FLOAT)0.55);
    gilVertex2f((FLOAT)0.95, (FLOAT)0.55);
    gilVertex2f((FLOAT)0.95, (FLOAT)0.95);
  }
  gilEnd();

  // After Right Click Line Drawn (Works for Concave and Self Intersect)
  gilBegin(GL_LINE_LOOP);
  {
    gilVertex2f((FLOAT)0.05, (FLOAT)0.45);
    gilVertex2f((FLOAT)0.05, (FLOAT)0.05);
    gilVertex2f((FLOAT)0.45, (FLOAT)0.05);
    gilVertex2f((FLOAT)0.45, (FLOAT)0.45);
  }
  gilEnd();

  // After Right Click Polygon Fill (Works for Convex only)
  gilPolygonMode(GL_FRONT, GL_FILL);
  gilPolygonMode(GL_BACK, GL_FILL);
  gilBegin(GL_POLYGON);
  {
    gilVertex2f((FLOAT)0.55, (FLOAT)0.45);
    gilVertex2f((FLOAT)0.55, (FLOAT)0.05);
    gilVertex2f((FLOAT)0.95, (FLOAT)0.05);
    gilVertex2f((FLOAT)0.95, (FLOAT)0.45);
  }
  gilEnd();

//
// END  : ASSIGNMENT 1: Rendering Code
///////////////////////////////////////////////////

  Togl_SwapBuffers(m_pTogl);

  t.Stop();
  SetUpdateTime(CSLIDEWorld::GetUpdateTime());
  SetRenderTime(t.GetSeconds());
}

VOID CWindow::HandleMouseButton(INT_32 iButton, INT_32 iState, INT_32 iX, INT_32 iY)
{
  DisplayEvent(eltCallBack, "Entering HandleMouseButton(%ld, %ld, %3ld, %3ld)\n", 
	       iButton, iState, iX, iY);

  BOOL b, bPostRedisplay;
  FLOAT fX, fY;
  CListIterInt iter;

  bPostRedisplay = FALSE;

  // Change Pixel coordinates into Normalize coords (0,0) -> (1,1)
  VPToNorm(fX, fY, iX, iY);

  if ( iState == GLSI_DOWN )
    {
///////////////////////////////////////////////////////
// START: ASSIGNMENT 1: Add data structure code here
// 
      // Fill in the appropriate actions and remember to set
      // bPostRedisplay = TRUE if you want to redraw any changes
      // to your data structures.

      switch( iButton ) 
	{
	case GLSI_LEFT_BUTTON:
	  DisplayEvent(eltAction, "L down  [%3ld %3ld]\n", iX, iY);
	  break;
	  
	case GLSI_MIDDLE_BUTTON:
	  DisplayEvent(eltAction, "M down  [%3ld %3ld]\n", iX, iY);
	  break;
	
	case GLSI_RIGHT_BUTTON:
	  DisplayEvent(eltAction, "R down  [%3ld %3ld]\n", iX, iY);
	  HandleWrite();
	  exit(0);
	  break;
	}
// 
// END  : ASSIGNMENT 1: Add data structure code here
///////////////////////////////////////////////////////
    }
  else if ( iState == GLSI_UP )
    {
      switch( iButton ) 
	{
	case GLSI_LEFT_BUTTON:
	  DisplayEvent(eltAction, "L up    [%3ld %3ld]\n", iX, iY);
	  break;
	  
	case GLSI_MIDDLE_BUTTON:
	  DisplayEvent(eltAction, "M up    [%3ld %3ld]\n", iX, iY);
	  break;
	
	case GLSI_RIGHT_BUTTON:
	  DisplayEvent(eltAction, "R up    [%3ld %3ld]\n", iX, iY);
	  break;
	}
    }

  if ( bPostRedisplay )
    {
      PostRedisplay();
    }
}

VOID CWindow::HandleMouseMotion(INT_32 iX, INT_32 iY)
{
  DisplayEvent(eltCallBack, "Entering HandleMouseMotion(%3ld, %3ld)\n", 
	       iX, iY);

  BOOL b, bPostRedisplay;
  FLOAT fX, fY;
  CListIterInt iter;

  bPostRedisplay = FALSE;

  // Change Pixel coordinates into Normalize coords (0,0) -> (1,1)
  VPToNorm(fX, fY, iX, iY);

  //sprintf(gpcBuffer, "%f", MIN(1.0, MAX(fX, 0.0)));
  //Tcl_SetVar(m_pInterp, "SLF_MOUSE_X", gpcBuffer, TCL_GLOBAL_ONLY);
  //sprintf(gpcBuffer, "%f", MIN(1.0, MAX(fY, 0.0)));
  //Tcl_SetVar(m_pInterp, "SLF_MOUSE_Y", gpcBuffer, TCL_GLOBAL_ONLY);

  if ( bPostRedisplay )
    {
      PostRedisplay();
    }
}

VOID CWindow::HandleKeyboard(INT_32 iKey, INT_32 iState, INT_32 iX, INT_32 iY)
{
  DisplayEvent(eltCallBack, "Entering HandleKeyboard('%c', %3ld, %3ld)\n",
	       (CHAR)iKey, iX, iY);

  BOOL bPostRedisplay;
  FLOAT fX, fY;

  bPostRedisplay = FALSE;

  // Change Pixel coordinates into Normalize coords (0,0) -> (1,1)
  VPToNorm(fX, fY, iX, iY);

  if ( iState != GLSI_DOWN )
    {
      return;
    }

  switch ( iKey )
    {
    case GLSI_KEY_ESCAPE:  // Escape Key
      exit(1);
      break;

    case 'e':
      SetDisplayEvents(!GetDisplayEvents());
      break;

    case 'w':
      HandleWrite();
      break;

    default:
      DisplayEvent(eltAction, "Key '%c' [%3ld %3ld]\n", (CHAR)iKey, iX, iY);
      break;
    }

  if ( bPostRedisplay )
    {
      PostRedisplay();
    }
}

VOID CWindow::HandleWrite()
{
  INT_32 iCode;
  UINT_32 uiLength;

  sprintf(gpcBuffer, "tk_getSaveFile -defaultextension .slf -filetypes {{\"SLIDE Files\" {.slf}}}");
  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
  ASSERT( iCode == TCL_OK );

  uiLength = strlen(m_pInterp->result);
  if ( uiLength == 0 )
    {
      // The user hit Cancel
      return;
    }

  if ( m_pcOutputFile != NULL )
    {
      delete []m_pcOutputFile;
      m_pcOutputFile = NULL;
    }
  ASSERT( m_fpOutputFile == NULL );

  ASSERT( m_pInterp->result != NULL );
  m_pcOutputFile = new CHAR[uiLength + 1];
  ASSERT( m_pcOutputFile != NULL );
  strcpy(m_pcOutputFile, m_pInterp->result);

  while ( m_fpOutputFile == NULL )
    {
      m_fpOutputFile = fopen(m_pcOutputFile, "w");
      if ( m_fpOutputFile == NULL )
	{
	  sprintf(gpcBuffer, "tk_messageBox -icon error -message \"File \\\"%s\\\" cannot be written.\" -type retrycancel -default retry", m_pcOutputFile);
	  iCode = Tcl_Eval(m_pInterp, gpcBuffer); 
	  ASSERT( iCode == TCL_OK );
	  
	  if ( strcmp(m_pInterp->result, "cancel") == 0 )
	    {
	      // Cancel
	      return;
	    }
	}
    }

  WriteSLIDE(m_fpOutputFile);

  if ( m_fpOutputFile != stdout )
    {
      ASSERT( m_fpOutputFile != NULL );
      fclose(m_fpOutputFile);
    }

  m_fpOutputFile = NULL;
}

VOID CWindow::SetDisplayEvents(BOOL b)
{
  m_bDisplayEvents = b;
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftDisplayEvents], b);
}

VOID CWindow::SetSecondsPerFrameRequested(DOUBLE dSecondsPerFrame)
{
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftFramesPerSecondRequested],
		     (dSecondsPerFrame == 0.0) ? (0.0) : (1.0/dSecondsPerFrame));
}

VOID CWindow::SetUpdateTime(DOUBLE dUpdateTime)
{
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftUpdateTime], dUpdateTime);
}

VOID CWindow::SetRenderTime(DOUBLE dRenderTime)
{
  Oot_SetObjectField(m_pOotObject, m_cppcOotFields[ftRenderTime], dRenderTime);
}

VOID CWindow::PostRedisplay()
{
  if ( !m_bRedisplayPending )
    { 
      ASSERT( m_pTogl != NULL );
      Togl_PostRedisplay(m_pTogl);
      m_bRedisplayPending = TRUE;
    }
}

VOID CWindow::DisplayEvent(EventLevelType eEventLevel, CHAR *pcFormat, ...)
{
  if ( m_bDisplayEvents && eEventLevel <= m_eEventLevel )
    {
      va_list vaArgs;

      va_start(vaArgs, pcFormat);
      vfprintf(stderr, pcFormat, vaArgs);
      va_end(vaArgs);
    }
}

//////////////////////////////////////////////////////////////////////
// Callback Functions
//

int CWindow::CreateOotClass(Tcl_Interp *pInterp)
{
  ASSERT( pInterp != NULL );

  m_pInterp = pInterp;

  m_pOotClass = Oot_CreateClass(m_pInterp, m_cpcOotClassName, 
				CWindow::CBHandleCreate, 
				CWindow::m_pOotMethods, 
				CWindow::m_cuiOotMethods,
				CWindow::m_cppcOotFields, 
				(UINT_32)(CWindow::ftSize));
  if ( m_pOotClass == NULL )
    {
      return TCL_ERROR;
    }

  return TCL_OK;
}
    
int CWindow::CBHandleCreate(Tcl_Interp *pInterp, OotClass *pClass, 
			    int argc, char ** /*argv*/)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pClass != NULL );

  if ( argc != 2 )
    {
      Tcl_SetResult(pInterp, "wrong args: window <name> <world>", NULL);
      return TCL_ERROR;
    }

  pCWindow = new CWindow;
  ASSERT( pCWindow != NULL );
  pCWindow->Init(NULL, 0, 0, NULL);
  /*
  if ( !pCWindow->Init() )
    {
      delete pCWindow;
      Tcl_SetResult(pInterp, "Error initializing window", NULL);
      return TCL_ERROR;      
    }
    */
  return TCL_OK;
}

VOID CWindow::CBHandleToglCreate(struct Togl *pTogl)
{
  ASSERT( pTogl != NULL );

  ASSERT( gpCWindowCurr != NULL );
  //printf("Togl client data = %s\n", gpCWindowCurr->GetSLIDEID());
  Togl_SetClientData(pTogl, (ClientData)gpCWindowCurr);
  gpCWindowCurr->SetTogl(pTogl);
  gpCWindowCurr = NULL;
}

VOID CWindow::CBHandleDisplay(struct Togl *pTogl)
{
  CWindow *pCWindow;

  ASSERT( pTogl != NULL );

  pCWindow = (CWindow*)Togl_GetClientData(pTogl);
  ASSERT( pCWindow != NULL );

  pCWindow->HandleDisplay();
}

VOID CWindow::CBHandleReshape(struct Togl *pTogl)
{
  CWindow *pCWindow;

  ASSERT( pTogl != NULL );

  pCWindow = (CWindow*)Togl_GetClientData(pTogl);
  ASSERT( pCWindow != NULL );

  pCWindow->HandleReshape(Togl_Width(pTogl), Togl_Height(pTogl));
}

VOID CWindow::CBHandleDestroy(struct Togl *pTogl)
{
  CWindow *pCWindow;

  ASSERT( pTogl != NULL );

  pCWindow = (CWindow*)Togl_GetClientData(pTogl);
  ASSERT( pCWindow != NULL );

  delete pCWindow;

  Togl_SetClientData(pTogl, NULL);
}

int CWindow::CBHandleMouseButton(Tcl_Interp *pInterp, OotObject *pObject,
				 INT_32 iButton, INT_32 iState,
				 INT_32 iX, INT_32 iY)
{
  CWindow *pCWindow;
  INT_32 iNewButton, iNewState;
  INT_32 iXVP, iYVP;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  switch ( iButton )
    {
    case 1:
      iNewButton = GLSI_LEFT_BUTTON;
      break;
    case 2:
      iNewButton = GLSI_MIDDLE_BUTTON;
      break;
    case 3:
      iNewButton = GLSI_RIGHT_BUTTON;
      break;
    default:
      ASSERT( FALSE );
      break;
    }
  
  iNewState = ( iState == 0 ) ? ( GLSI_DOWN ) : ( GLSI_UP );

  pCWindow->WinToVP(iXVP, iYVP, (INT_32)iX, (INT_32)iY);

  pCWindow->HandleMouseButton(iNewButton, iNewState, iXVP, iYVP);

  return TCL_OK;
}

int CWindow::CBHandleMouseMotion(Tcl_Interp *pInterp, OotObject *pObject,
				 INT_32 /*iButton*/, INT_32 iX, INT_32 iY)
{
  CWindow *pCWindow;
  INT_32 iXVP, iYVP;
  
  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->WinToVP(iXVP, iYVP, (INT_32)iX, (INT_32)iY);

  pCWindow->HandleMouseMotion(iXVP, iYVP);

  return TCL_OK;
}

int CWindow::CBHandleKeyboard(Tcl_Interp *pInterp, OotObject *pObject,
			      CHAR *pcKeyCode, INT_32 iState, 
			      INT_32 iX, INT_32 iY)
{
  CWindow *pCWindow;
  INT_32 iKey, iNewState;
  glsiKeyBinding *pKeyBinding;
  INT_32 iXVP, iYVP;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pKeyBinding = (glsiKeyBinding*)(m_pHashKeyboard->Peek(pcKeyCode));
  if ( pKeyBinding != NULL )
    {
      iKey = pKeyBinding->iValue;
      
      iNewState = ( iState == 0 ) ? ( GLSI_DOWN ) : ( GLSI_UP );
      pCWindow->WinToVP(iXVP, iYVP, (INT_32)iX, (INT_32)iY);
      
      pCWindow->HandleKeyboard(iKey, iNewState, iXVP, iYVP);
    }

  return TCL_OK;
}

int CWindow::CBHandleWrite(Tcl_Interp *pInterp, OotObject *pObject)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->HandleWrite();

  return TCL_OK;
}

int CWindow::CBSetDisplayEvents(Tcl_Interp *pInterp, OotObject *pObject, BOOL b)
{
  CWindow *pCWindow;

  ASSERT( pInterp != NULL && pObject != NULL );

  pCWindow = (CWindow*)(pObject->pvClientData);
  ASSERT( pCWindow != NULL );

  pCWindow->SetDisplayEvents(b);
  sprintf(gpcBuffer, "%ld", (UINT_32)(b));
  Tcl_AppendResult(pInterp, gpcBuffer, NULL);

  return TCL_OK;
}

